package pesto;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.*;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;

import org.apache.commons.io.FileUtils;
import org.aspectj.lang.JoinPoint;
import org.aspectj.lang.annotation.Before;
import org.opencv.imgproc.Imgproc;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.WebElement;
import org.openqa.selenium.support.ui.Select;

public aspect VisualLocatorGenerator {

	protected static String screenshotFolder;
	protected static String poSikuliFolder;
	protected static String lfn;
	protected static File locFile;

	protected static Properties configFile;

	static WebDriver d = null;
	static Map<String, LinkedList<String>> list = new HashMap<String, LinkedList<String>>();

	/**
	 * This advice starts at any Selenium WebDriver manage invocation. It sets up
	 * the directories in which PESTO saves the screenshots and all the files.
	 * 
	 * @param joinPoint
	 */
	@Before("call(* org.openqa.selenium.WebDriver.manage())")
	public void logBeforeWebDriverManage(JoinPoint joinPoint) {

		configFile = new Properties();

		try
		{
			configFile.load(VisualLocatorGenerator.class.getClassLoader().getResourceAsStream("config.properties"));

			screenshotFolder = configFile.getProperty("SCREENSHOTS_FOLDER");
			poSikuliFolder = configFile.getProperty("PO_SIKULI_FOLDER");
			lfn = configFile.getProperty("LOCATORS_FILE");
			locFile = new File(lfn);

		} catch (IOException e)
		{
			e.printStackTrace();
		}

		File theDir = new File(screenshotFolder);
		if (!theDir.exists())
		{
			System.out.println("creating directory " + screenshotFolder + "...");
			boolean result = theDir.mkdir();
			if (result)
			{
				System.out.println("...created!");
			}
		}

		if (!locFile.exists()) {
			System.out.println("creating file locators.txt " + "...");
			try
			{
				locFile.createNewFile();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}

		File poSikuliDir = new File(poSikuliFolder);

		if (!poSikuliDir.exists())
		{
			System.out.println("creating directory " + poSikuliFolder + "...");
			boolean result = poSikuliDir.mkdir();
			if (result)
			{
				System.out.println("...created!");
			}
		}

		WebDriver driver = (WebDriver) joinPoint.getTarget();
		d = driver;
	}

	// These advices capture all action calls on any WebElement and call the
	// Visual Locator Generator

	@Before("call(* org.openqa.selenium.WebElement.click())")
	public void clickInterceptor(JoinPoint joinPoint) {
		callVisualLocatorCreator(joinPoint, "CLICK");
	}

	@Before("call(* org.openqa.selenium.WebElement.sendKeys(..))")
	public void sendKeysInterceptor(JoinPoint joinPoint) {
		callVisualLocatorCreator(joinPoint, "SEND_KEYS");
	}

	@Before("call(* org.openqa.selenium.WebElement.getText())")
	public void getTextInterceptor(JoinPoint joinPoint) {
		callVisualLocatorCreator(joinPoint, "GET_TEXT");
	}

	@Before("call(* org.openqa.selenium.support.ui.Select.selectByVisibleText(..))")
	public void selectVisibleTextInterceptor(JoinPoint joinPoint) {
		callVisualLocatorCreator(joinPoint, "SELECT");
	}
	
	@Before("call(* org.openqa.selenium.support.ui.Select.selectByIndex(..))")
	public void selectIndexInterceptor(JoinPoint joinPoint) {
		callVisualLocatorCreator(joinPoint, "SELECT");
	}

	@Before("call(* org.openqa.selenium.WebElement.clear(..))")
	public void clearInterceptor(JoinPoint joinPoint) {
		callVisualLocatorCreator(joinPoint, "CLEAR");
	}
	
	public void callVisualLocatorCreator(JoinPoint joinPoint, String text)
	{
		configFile = new Properties();

		try
		{
			configFile.load(VisualLocatorGenerator.class.getClassLoader().getResourceAsStream("config.properties"));
		} catch (IOException e)
		{
			e.printStackTrace();
		}

		String isClickEnabled = configFile.getProperty(text);
		if (isClickEnabled.equals("y"))
		{
			System.out.println("[LOG] selenium.WebElement. " + text + " interception is enabled");
			visualLocatorCreator(joinPoint);
		}
		else
		{
			System.out.println("[LOG] selenium.WebElement. " + text + " interception is disabled");
		}

	}

	/**
	 * Proxy to the visual locator generator advice
	 * 
	 * @param joinPoint
	 */
	public void visualLocatorCreator(JoinPoint joinPoint) {
		VisualLocatorGenerator.logBeforeWebElement(joinPoint);
	}

	/**
	 * The heart of the Visual Locator Generator
	 * 
	 * @param joinPoint
	 */
	private static void logBeforeWebElement(JoinPoint joinPoint) {

		WebElement e = null;
		Select s = null;
		String var = "";

		if (joinPoint.getTarget() instanceof Select) {
			s = (Select) joinPoint.getTarget();
			e = (WebElement) s.getOptions().get(0);
			var = extractLocator(s.getOptions().get(0).toString());
		} else {
			e = (WebElement) joinPoint.getTarget();
			var = extractLocator(joinPoint.getTarget().toString());
		}

		String webElementImageName = joinPoint.getSourceLocation().getFileName().replace(".java", "") + "/" + joinPoint.getSourceLocation().getLine();

		// find the PO name and create the new PO file, if it doesn't exist
		String po = joinPoint.getSourceLocation().getFileName().replace(".java", "");
		String poM = poSikuliFolder + po.concat(".java");
		File modifiedPo = new File(poM);

		if (!modifiedPo.exists()) {
			try {
				modifiedPo.createNewFile();
			} catch (IOException e1) {
				e1.printStackTrace();
			}
		}

		String loc = po + "-" + var;

		// multistate
		if (list.containsKey(loc)) {
			Iterator<String> itr = list.get(loc).iterator();
			while (itr.hasNext()) {
				String el = itr.next();
				webElementImageName = webElementImageName.replace('/', '-');
				if (!el.toLowerCase().contains(webElementImageName.toLowerCase())) {
					System.err.println("[WARNING] MULTISTATE ELEMENT: " + loc);
					webElementImageName = webElementImageName.replace('-', '/');
				}
				webElementImageName = webElementImageName.replace('-', '/');
			}
		}

		webElementImageName = webElementImageName + "-" + System.currentTimeMillis();

		String pathScreenShot = "";
		String pathVisualLocator = "";

		List<String> images = new LinkedList<String>();
		int scale = 4; // nemro passi per arrivare al locator pi grande partendo
										// dalla size dell'elemento
		boolean matchUnivocal = false;
		try {

			do {
				images = UtilsOpenCv.shoot(d, e, webElementImageName, scale);
				pathScreenShot = images.get(0);
				pathVisualLocator = images.get(1);
				matchUnivocal = UtilsOpenCv.isUniqueMatch(pathScreenShot, pathVisualLocator, locFile.getPath(), Imgproc.TM_CCOEFF_NORMED);
				scale--;
			} while (matchUnivocal == false);
		} catch (IOException | InterruptedException e1)
		{
			e1.printStackTrace();
		}

		// ////////////////////////////////////////////////////////
		String dirName = images.get(1).substring(0, images.get(1).lastIndexOf("/"));

	  // loca contiene il numero di linea quindi i cloni con numero di linea diversi sono lasciati
		//String loca = webElementImageName.replace('/', '-').substring(0, webElementImageName.lastIndexOf("-"));
		
	  // loca NON contiene il numero di linea quindi si guardano i cloni in tutta la pagina
		String loca = webElementImageName.substring(0, webElementImageName.indexOf("/"));
		
		File dir = new File(dirName);
		File[] foundFiles = dir.listFiles(new FilenameFilter() {
			public boolean accept(File dir, String name) {
				return name.startsWith(loca);
				
			}
		});
		
		if (pathVisualLocator.contains("IndexPage-67"))
			System.out.println("ciao");

		try {
			File currentLocatorImage = new File(pathVisualLocator);
			for (File file : foundFiles) {

				if (file.getAbsolutePath().endsWith("All.png")) 
				{
					// discard complete screenshots
				}
				else if (file.getAbsolutePath().contentEquals(currentLocatorImage.getAbsolutePath())) 
				{
					// self discard
				}
				else 
				{
					//leviamo solo i cloni dello stesso numero di linea, quelli tra numeri di linea diversi li lasciamo

					System.out.println("[LOG]   comparing: "+file.getAbsolutePath());
					System.out.println("[LOG]   with:      "+currentLocatorImage.getAbsolutePath());

					boolean areEqual = FileUtils.contentEquals(file, currentLocatorImage);

					System.out.println("[LOG]   equal?     "+ areEqual);

					if (areEqual)
					{
						System.out.println("a" + pathVisualLocator);
						System.out.println("b" + pathVisualLocator.substring(0, pathVisualLocator.lastIndexOf("/"))
								+ file.getAbsolutePath().substring(file.getAbsolutePath().lastIndexOf("\\")).replace('\\', '/')
								);
						pathVisualLocator = pathVisualLocator.substring(0, pathVisualLocator.lastIndexOf("/"))
								+ file.getAbsolutePath().substring(file.getAbsolutePath().lastIndexOf("\\")).replace('\\', '/');

						currentLocatorImage.delete();
						File currentScreenshotImage = new File(pathScreenShot);
						currentScreenshotImage.delete();
						break;
					}
				}
			}
			System.out.println("\n");
		} catch (IOException e1)
		{
			e1.printStackTrace();
		}

		// ////////////////////////////////////////////////////////

		// loc = "IndexPage-xpath="html/body/..."
		if (list.get(loc) == null) {
			LinkedList<String> l = new LinkedList<String>();
			l.add(pathVisualLocator);
			list.put(loc, l);
		}
		else
		{
			if (!list.get(loc).contains(pathVisualLocator)) {
				list.get(loc).add(pathVisualLocator);
			}
			list.put(loc, list.get(loc));
		}

		System.out.println(po + " " + var + " " + images.get(1));
		System.out.println(po + " " + var + " " + pathVisualLocator);
	}

	/**
	 * Auxiliary method to extract the locator from the @FindBy Selenium WebDriver
	 * annotation.
	 * 
	 * @param s
	 *          the string representing the annotation
	 * @return
	 */
	private static String extractLocator(String s) {

		String var = s;
		var = (String) var.subSequence(var.indexOf(" -> ") + 4, var.length() - 1);

		if (var.contains("->")) {
			var = (String) var.subSequence(0, var.indexOf(" -> "));
			var = (String) var.substring(0, var.length() - 2);
		}

		var = var.replace(": ", "=\"");
		var = var + "\"";

		// uniform locators notations to Selenium WebDriver's
		if (var.contains("css selector")) {
			var = var.replace("css selector", "css");
		} else if (var.contains("link text")) {
			var = var.replace("link text", "linkText");
		} else if (var.contains("xpath")) {
			;
		} else if (var.contains("name")) {
			;
		}
		return var;
	}

	/**
	 * This advice starts at any Selenium WebDriver quit invocation. It saves all
	 * the locators mapping on the filesystem on a dedicated file
	 * 
	 * @param joinPoint
	 */
	@Before("call(* org.openqa.selenium.WebDriver.quit())")
	public void logBeforeWebDriverQuit(JoinPoint joinPoint) {

		// at the end of the test case, save the all the locators in the file
		BufferedWriter writer;
		try {
			writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(lfn, true), "utf-8"));
			BufferedReader br = new BufferedReader(new FileReader(lfn));

			for (Entry<String, LinkedList<String>> entry : list.entrySet()) {
				String key = entry.getKey();
				LinkedList<String> value = entry.getValue();
				String fileLocators = FileUtils.readFileToString(locFile);

				for (int i = 0; i < value.size(); i++) {
					if (!fileLocators.contains(key + " -> " + value.get(i))) {
						writer.append(key + " -> " + value.get(i) + "\n");
					}
				}
			}
			br.close();
			writer.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
